from __future__ import unicode_literals


from commonmark.render.renderer import Renderer


class ReStructuredTextRenderer(Renderer):
    """
    Render reStructuredText from Markdown

    Example:

    .. code:: python

        import commonmark

        parser = commonmark.Parser()
        ast = parser.parse('Hello `inline code` example')

        renderer = commonmark.ReStructuredTextRenderer()
        rst = renderer.render(ast)
        print(rst)  # Hello ``inline code`` example
    """
    def __init__(self, indent_char=' '):
        self.indent_char = indent_char
        self.indent_length = 0

    def lit(self, s):
        if s == '\n':
            indent = ''  # Avoid whitespace if we're just adding a newline
        elif self.last_out != '\n':
            indent = ''  # Don't indent if we're in the middle of a line
        else:
            indent = self.indent_char * self.indent_length

        return super(ReStructuredTextRenderer, self).lit(indent + s)

    def cr(self):
        self.lit('\n')

    def indent_lines(self, literal, indent_length=4):
        indent = self.indent_char * indent_length
        new_lines = []

        for line in literal.splitlines():
            new_lines.append(indent + line)

        return '\n'.join(new_lines)

    # Nodes

    def document(self, node, entering):
        pass

    def softbreak(self, node, entering):
        self.cr()

    def linebreak(self, node, entering):
        self.cr()
        self.cr()

    def text(self, node, entering):
        self.out(node.literal)

    def emph(self, node, entering):
        self.out('*')

    def strong(self, node, entering):
        self.out('**')

    def paragraph(self, node, entering):
        if node.parent.t == 'item':
            pass
        else:
            self.cr()

    def link(self, node, entering):
        if entering:
            self.out('`')
        else:
            self.out(' <%s>`_' % node.destination)

    def image(self, node, entering):
        directive = '.. image:: ' + node.destination

        if entering:
            self.out(directive)
            self.cr()
            self.indent_length += 4
            self.out(':alt: ')
        else:
            self.indent_length -= 4

    def code(self, node, entering):
        self.out('``')
        self.out(node.literal)
        self.out('``')

    def code_block(self, node, entering):
        directive = '.. code::'
        language_name = None

        info_words = node.info.split() if node.info else []
        if len(info_words) > 0 and len(info_words[0]) > 0:
            language_name = info_words[0]

        if language_name:
            directive += ' ' + language_name

        self.cr()
        self.out(directive)
        self.cr()
        self.cr()
        self.out(self.indent_lines(node.literal))
        self.cr()

    def list(self, node, entering):
        if entering:
            self.cr()

    def item(self, node, entering):
        tagname = '* ' if node.list_data['type'] == 'bullet' else '#. '

        if entering:
            self.out(tagname)
            self.indent_length += len(tagname)
        else:
            self.indent_length -= len(tagname)
            self.cr()

    def block_quote(self, node, entering):
        if entering:
            self.indent_length += 4
        else:
            self.indent_length -= 4

    def heading(self, node, entering):
        heading_chars = [
            '#',
            '*',
            '=',
            '-',
            '^',
            '"'
        ]

        try:
            heading_char = heading_chars[node.level-1]
        except IndexError:
            # Default to the last level if we're in too deep
            heading_char = heading_chars[-1]

        heading_length = len(node.first_child.literal)
        banner = heading_char * heading_length

        if entering:
            self.cr()
        else:
            self.cr()
            self.out(banner)
            self.cr()



from __future__ import absolute_import, unicode_literals

import re
import sys

try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote

if sys.version_info >= (3, 0):
    if sys.version_info >= (3, 4):
        import html
        HTMLunescape = html.unescape
    else:
        from .entitytrans import _unescape
        HTMLunescape = _unescape
else:
    from commonmark import entitytrans
    HTMLunescape = entitytrans._unescape

ENTITY = '&(?:#x[a-f0-9]{1,6}|#[0-9]{1,7}|[a-z][a-z0-9]{1,31});'

TAGNAME = '[A-Za-z][A-Za-z0-9-]*'
ATTRIBUTENAME = '[a-zA-Z_:][a-zA-Z0-9:._-]*'
UNQUOTEDVALUE = "[^\"'=<>`\\x00-\\x20]+"
SINGLEQUOTEDVALUE = "'[^']*'"
DOUBLEQUOTEDVALUE = '"[^"]*"'
ATTRIBUTEVALUE = "(?:" + UNQUOTEDVALUE + "|" + SINGLEQUOTEDVALUE + \
    "|" + DOUBLEQUOTEDVALUE + ")"
ATTRIBUTEVALUESPEC = "(?:" + "\\s*=" + "\\s*" + ATTRIBUTEVALUE + ")"
ATTRIBUTE = "(?:" + "\\s+" + ATTRIBUTENAME + ATTRIBUTEVALUESPEC + "?)"
OPENTAG = "<" + TAGNAME + ATTRIBUTE + "*" + "\\s*/?>"
CLOSETAG = "</" + TAGNAME + "\\s*[>]"
HTMLCOMMENT = '<!---->|<!--(?:-?[^>-])(?:-?[^-])*-->'
PROCESSINGINSTRUCTION = "[<][?].*?[?][>]"
DECLARATION = "<![A-Z]+" + "\\s+[^>]*>"
CDATA = '<!\\[CDATA\\[[\\s\\S]*?\\]\\]>'
HTMLTAG = "(?:" + OPENTAG + "|" + CLOSETAG + "|" + HTMLCOMMENT + "|" + \
    PROCESSINGINSTRUCTION + "|" + DECLARATION + "|" + CDATA + ")"
reHtmlTag = re.compile('^' + HTMLTAG, re.IGNORECASE)
reBackslashOrAmp = re.compile(r'[\\&]')
ESCAPABLE = '[!"#$%&\'()*+,./:;<=>?@[\\\\\\]^_`{|}~-]'
reEntityOrEscapedChar = re.compile(
    '\\\\' + ESCAPABLE + '|' + ENTITY, re.IGNORECASE)
XMLSPECIAL = '[&<>"]'
reXmlSpecial = re.compile(XMLSPECIAL)


def unescape_char(s):
    if s[0] == '\\':
        return s[1]
    else:
        return HTMLunescape(s)


def unescape_string(s):
    """Replace entities and backslash escapes with literal characters."""
    if re.search(reBackslashOrAmp, s):
        return re.sub(
            reEntityOrEscapedChar,
            lambda m: unescape_char(m.group()),
            s)
    else:
        return s


def normalize_uri(uri):
    try:
        return quote(uri.encode('utf-8'), safe=str(';/@:+?=&()%#*,'))
    except UnicodeDecodeError:
        # Python 2 also throws a UnicodeDecodeError, complaining about
        # the width of the "safe" string. Removing this parameter
        # solves the issue, but yields overly aggressive quoting, but we
        # can correct those errors manually.
        s = quote(uri.encode('utf-8'))
        s = re.sub(r'%40', '@', s)
        s = re.sub(r'%3A', ':', s)
        s = re.sub(r'%2B', '+', s)
        s = re.sub(r'%3F', '?', s)
        s = re.sub(r'%3D', '=', s)
        s = re.sub(r'%26', '&', s)
        s = re.sub(r'%28', '(', s)
        s = re.sub(r'%29', ')', s)
        s = re.sub(r'%25', '%', s)
        s = re.sub(r'%23', '#', s)
        s = re.sub(r'%2A', '*', s)
        s = re.sub(r'%2C', ',', s)
        return s


UNSAFE_MAP = {
    '&': '&amp;',
    '<': '&lt;',
    '>': '&gt;',
    '"': '&quot;',
}


def replace_unsafe_char(s):
    return UNSAFE_MAP.get(s, s)


def escape_xml(s):
    if s is None:
        return ''
    if re.search(reXmlSpecial, s):
        return re.sub(
            reXmlSpecial,
            lambda m: replace_unsafe_char(m.group()),
            s)
    else:
        return s
